from typing import Literal

import click

from pytidb.ext.mcp.server import mcp, log


@click.command()
@click.option(
    "--transport",
    type=click.Choice(["stdio", "sse"]),
    default="stdio",
    help="Transport type",
)
def main(transport: Literal["stdio", "sse"] = "stdio"):
    log.info("Starting tidb mcp server...")
    mcp.run(transport=transport)



from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Union,
    Tuple,
    Sequence,
    TypeVar,
    Generic,
)

from pydantic import BaseModel, Field
from sqlalchemy import Row, Select, asc, desc, literal, select, and_, text
from pytidb.functions import fts_match_word
from pytidb.rerankers.base import BaseReranker
from pytidb.schema import DistanceMetric, QueryBundle, VectorDataType, TableModel
from pytidb.utils import (
    build_filter_clauses,
    check_text_column,
    check_vector_column,
    get_row_id_from_row,
    merge_result_rows,
)
from pytidb.logger import logger


if TYPE_CHECKING:
    from pytidb.table import Table
    from pandas import DataFrame


SearchType = Literal["vector", "fulltext", "hybrid"]


DISTANCE_LABEL = "_distance"
MATCH_SCORE_LABEL = "_match_score"
SCORE_LABEL = "_score"
ROW_ID_LABEL = "_tidb_rowid"

T = TypeVar("T", bound=TableModel)


class SearchResult(BaseModel, Generic[T]):
    hit: T
    distance: Optional[float] = Field(
        description="The distance between the query vector and the vectors in the table.",
        default=None,
    )
    match_score: Optional[float] = Field(
        description="The match score between the query text and the text in the table.",
        default=None,
    )
    score: Optional[float] = Field(
        description="The score of the search result.",
        default=None,
    )

    def __getattr__(self, item: str):
        if hasattr(self.hit, item):
            return getattr(self.hit, item)
        raise AttributeError(
            f"'{self.__class__.__name__}' object has no attribute '{item}'"
        )

    @property
    def similarity_score(self) -> float:
        if self.distance is not None:
            return 1 - self.distance
        else:
            return None


class SearchQuery:
    def __init__(
        self,
        table: "Table",
        search_type: SearchType = "vector",
        query: Optional[Union[VectorDataType, str, QueryBundle]] = None,
    ):
        # Table.
        self._table = table
        self._sa_table = table._sa_table
        self._client = table.client
        self._columns = table._columns
        self._vector_column = table.vector_column
        self._text_column = table.text_column

        # Query.
        self._search_type = search_type
        self._query = query
        if isinstance(query, dict):
            self._query_vector = query["query_vector"]
            self._query_text = query["query_text"]
        else:
            self._query_vector = query if isinstance(query, list) else None
            self._query_text = query if isinstance(query, str) else None

        # Vector search.
        self._distance_metric = DistanceMetric.COSINE
        self._distance_threshold = None
        self._distance_lower_bound = None
        self._distance_upper_bound = None
        self._filters = None
        self._num_candidate = 20

        # Reranker.
        self._reranker = None
        self._rerank_field_name = None

        # Miscellaneous.
        self._debug = False
        self._limit = None

    def vector(self, query_vector: VectorDataType):
        self._query_vector = query_vector
        return self

    def text(self, query_text: str):
        self._query_text = query_text
        return self

    def vector_column(self, column_name: str):
        self._vector_column = check_vector_column(self._columns, column_name)
        return self

    def text_column(self, column_name: str):
        self._text_column = check_text_column(self._columns, column_name)
        return self

    def distance_metric(self, metric: DistanceMetric) -> "SearchQuery":
        self._distance_metric = metric
        return self

    def distance_threshold(self, threshold: Optional[float] = None) -> "SearchQuery":
        self._distance_threshold = threshold
        return self

    def distance_range(
        self, lower_bound: float = 0, upper_bound: float = 1
    ) -> "SearchQuery":
        self._distance_lower_bound = lower_bound
        self._distance_upper_bound = upper_bound
        return self

    def num_candidate(self, num_candidate: int) -> "SearchQuery":
        self._num_candidate = num_candidate
        return self

    def filter(self, filters: Optional[Dict[str, Any]] = None) -> "SearchQuery":
        self._filters = filters
        return self

    def limit(self, k: int) -> "SearchQuery":
        self._limit = k
        return self

    def debug(self, flag: bool = True) -> "SearchQuery":
        self._debug = flag
        return self

    def rerank(
        self, reranker: BaseReranker, rerank_field: Optional[str] = None
    ) -> "SearchQuery":
        """
        Rerank the search results.

        Reranker can be used for multiple-way (multi-vectors, hybrid) search result fusion or
        improve the quality of vector search results.

        Args:
            reranker: The reranker to use.
            rerank_field: The field to rerank on.
        """
        self._reranker = reranker
        self._rerank_field_name = rerank_field
        return self

    def _build_vector_query(self) -> Select:
        num_candidate = self._num_candidate if self._num_candidate else self._limit * 10

        if self._vector_column is None:
            if len(self._table.vector_columns) == 0:
                raise ValueError(
                    "vector column is not found, but is required for vector search"
                )
            elif len(self._table.vector_columns) >= 1:
                raise ValueError(
                    "more than two vector columns, please choice one through .vector_column()"
                )
            else:
                vector_column = self._table.vector_columns[0]
        else:
            vector_column = self._vector_column

        # Auto embedding
        if self._query_vector is None:
            if vector_column.name not in self._table.vector_field_configs:
                raise ValueError(
                    "query should be a vector, because the vector column didn't configure the embed_fn parameter"
                )

            config = self._table.vector_field_configs[vector_column.name]
            self._query_vector = config["embed_fn"].get_query_embedding(
                self._query_text
            )

        # Distance metric.
        if self._distance_metric == DistanceMetric.L2:
            distance_column = vector_column.l2_distance(self._query_vector).label(
                DISTANCE_LABEL
            )
        else:
            distance_column = vector_column.cosine_distance(self._query_vector).label(
                DISTANCE_LABEL
            )

        # Inner query for ANN search
        table_model = self._table.table_model
        columns = table_model.__table__.c
        inner_select_columns = list(columns)
        if self._sa_table.primary_key is None:
            inner_select_columns.insert(0, text("_tidb_rowid"))

        subquery_stmt = (
            select(
                *columns,
                distance_column,
            )
            .order_by(asc(DISTANCE_LABEL))
            .limit(num_candidate)
        )

        # Distance range.
        having = []
        if self._distance_lower_bound and self._distance_upper_bound:
            having.append(distance_column >= self._distance_lower_bound)
            having.append(distance_column <= self._distance_upper_bound)

        # Distance threshold.
        if self._distance_threshold:
            having.append(distance_column <= self._distance_threshold)

        if len(having) > 0:
            subquery_stmt = subquery_stmt.having(and_(*having))

        subquery = subquery_stmt.subquery("candidates")

        # Main query with metadata filters
        outer_select_columns = list(subquery.c)
        if self._sa_table.primary_key is None:
            outer_select_columns.insert(0, text("_tidb_rowid"))

        stmt = select(
            *outer_select_columns,
            (1 - subquery.c[DISTANCE_LABEL]).label(SCORE_LABEL),
        )

        if self._filters is not None:
            filter_clauses = build_filter_clauses(
                self._filters, subquery.c, table_model
            )
            stmt = stmt.filter(*filter_clauses)

        stmt = stmt.order_by(asc(DISTANCE_LABEL)).limit(self._limit)

        # Debug.
        if self._debug:
            db_engine = self._table.db_engine
            table_name = self._table.table_name
            compiled_sql = stmt.compile(
                dialect=db_engine.dialect, compile_kwargs={"literal_binds": True}
            )
            logger.info(
                f"Build vector search query on table <{table_name}>:\n{compiled_sql}"
            )

        return stmt

    def _build_fulltext_query(self) -> Select:
        if self._query_text is None:
            raise ValueError(
                "query string is required for fulltext search, please specify it through "
                ".text('<your query string>')"
            )

        if self._text_column is None:
            if len(self._table.text_columns) == 0:
                raise ValueError(
                    "no text column found in the table, fulltext search cannot be executed"
                )
            elif len(self._table.text_columns) >= 1:
                raise ValueError(
                    "more than two text columns in the table, need to specify one through"
                    ".text_column('<your text column name>')"
                )
            else:
                text_column = self._table.text_columns[0]
        else:
            text_column = self._text_column

        table_model = self._table.table_model
        columns = table_model.__table__.c
        select_columns = list(columns)
        if self._sa_table.primary_key is None:
            select_columns.insert(0, text("_tidb_rowid"))

        # TODO: support return score after fts_match_word is supported.
        table_name = self._table.table_name
        stmt = select(
            *select_columns,
            literal(None).label(MATCH_SCORE_LABEL),
            literal(None).label(SCORE_LABEL),
        ).filter(fts_match_word(self._query_text, text_column))

        if self._filters is not None:
            filter_clauses = build_filter_clauses(self._filters, columns, table_model)
            stmt = stmt.filter(*filter_clauses)

        stmt = stmt.order_by(desc(fts_match_word(self._query_text, text_column))).limit(
            self._limit
        )

        # Debug.
        if self._debug:
            db_engine = self._table.db_engine
            table_name = self._table.table_name
            compiled_sql = stmt.compile(
                dialect=db_engine.dialect, compile_kwargs={"literal_binds": True}
            )
            logger.info(
                f"Build fulltext search query on table <{table_name}>:\n{compiled_sql}"
            )

        return stmt

    def _execute_query(self) -> Tuple[List[str], List[Any]]:
        if self._limit is None:
            raise ValueError(
                "limit is required for search, please specify it through " ".limit(n)"
            )

        if self._search_type == "vector":
            return self._exec_vector_query()
        elif self._search_type == "fulltext":
            return self._exec_fulltext_query()
        elif self._search_type == "hybrid":
            return self._exec_hybrid_query()
        else:
            raise ValueError(
                f"invalid search type: {self._search_type}, allowed search types are "
                "`vector`, `fulltext`, and `hybrid`"
            )

    def _exec_vector_query(self) -> Tuple[List[str], List[Row]]:
        with self._client.session() as db_session:
            vector_query = self._build_vector_query()
            result = db_session.execute(vector_query)
            keys = result.keys()
            rows = result.fetchall()

            # Apply reranker to improve the accuracy of vector search results. (Optional)
            if self._reranker is not None:
                rows = self._rerank_result_set(rows)

            return keys, rows

    def _exec_fulltext_query(self) -> Tuple[List[str], List[Row]]:
        with self._client._db_engine.connect() as conn:
            query = self._build_fulltext_query()
            result = conn.execute(query)
            keys = result.keys()
            rows = result.fetchall()

        # Apply reranker to improve the accuracy of fulltext search results. (Optional)
        if self._reranker is not None:
            rows = self._rerank_result_set(rows)

        return keys, rows

    def _exec_hybrid_query(self) -> Tuple[List[str], List[Row]]:
        with self._client.session() as db_session:
            vs_query = self._build_vector_query()
            vs_result = db_session.execute(vs_query)
            vs_rows = vs_result.fetchall()

            fts_query = self._build_fulltext_query()
            fts_result = db_session.execute(fts_query)
            fts_rows = fts_result.fetchall()

            # Merge the rows from vector search and fulltext search.
            def get_row_id(row: Row) -> Optional[int]:
                return get_row_id_from_row(row, self._sa_table)

            keys, rows = merge_result_rows(vs_rows, fts_rows, get_row_id)

            # Apply reranker to rerank the merged result set. (Optional)
            if self._reranker is not None:
                rows = self._rerank_result_set(rows)
            else:
                # Sort the rows by score.
                rows = sorted(
                    rows, key=lambda row: row._mapping[SCORE_LABEL] or 0, reverse=True
                )

            return keys, rows[: self._limit]

    def _rerank_result_set(self, rows: List[Row]) -> List[Row]:
        """
        Rerank the search results.

        Args:
            rows: The rows to rerank.

        Returns:
            The reranked rows.
        """
        rerank_field_name = self._get_rerank_field_name()

        if self._query_text is None:
            raise ValueError(
                "query text is required for reranker, please specify it through "
                ".text('<your query string>')"
            )

        documents = [row._mapping[rerank_field_name] for row in rows]
        reranked_results = self._reranker.rerank(
            self._query_text, documents, self._limit
        )
        reranked_rows = []
        for item in reranked_results:
            row = rows[item.index]
            score_index = row._key_to_index["_score"]
            _data = list(row._tuple())
            # Replace the score with the reranked score.
            _data[score_index] = item.relevance_score
            reranked_rows.append(
                Row(
                    row._parent,
                    None,
                    row._key_to_index,
                    tuple(_data),
                )
            )
        return reranked_rows

    def _get_rerank_field_name(self) -> str:
        if self._rerank_field_name is not None:
            return self._rerank_field_name

        if self._search_type in ["vector", "hybrid"]:
            if self._vector_column is not None:
                vector_field = self._table.vector_field_configs[
                    self._vector_column.name
                ]
                return vector_field["source_field_name"]

        if self._search_type == "fulltext":
            if self._text_column is not None:
                return self._text_column

        raise ValueError(
            "Please specify the rerank field name through .rerank(reranker, rerank_field_name)"
        )

    def to_rows(self) -> Sequence[Any]:
        _, rows = self._execute_query()
        return rows

    def to_list(self) -> List[dict]:
        keys, rows = self._execute_query()
        results = [dict(zip(keys, row)) for row in rows]
        return results

    def to_pydantic(self, with_score: Optional[bool] = True) -> List[BaseModel]:
        table_model = self._table.table_model

        _, rows = self._execute_query()
        results = []
        for row in rows:
            values: Dict[str, Any] = dict(row._mapping)
            distance = values.pop(DISTANCE_LABEL) if DISTANCE_LABEL in values else None
            match_score = (
                values.pop(MATCH_SCORE_LABEL) if MATCH_SCORE_LABEL in values else None
            )
            score = values.pop(SCORE_LABEL) if SCORE_LABEL in values else None
            hit = table_model.model_validate(values)

            if not with_score:
                results.append(hit)
            else:
                results.append(
                    SearchResult(
                        distance=distance,
                        match_score=match_score,
                        score=score,
                        hit=hit,
                    )
                )

        return results

    def to_pandas(self) -> "DataFrame":
        try:
            import pandas as pd
        except Exception:
            raise ImportError(
                "Failed to import pandas, please install it with `pip install pandas`"
            )

        keys, rows = self._execute_query()
        return pd.DataFrame(rows, columns=keys)



import logging
import os
import re

from dotenv import load_dotenv
from contextlib import asynccontextmanager
from typing import AsyncIterator, Optional
from mcp.server.fastmcp import FastMCP, Context
from dataclasses import dataclass

from pydantic import MySQLDsn

from pytidb import TiDBClient
from pytidb.utils import TIDB_SERVERLESS_HOST_PATTERN

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
log = logging.getLogger("tidb_mcp_server")

# Load environment variables
load_dotenv()

# Constants
TIDB_SERVERLESS_USERNAME_PATTERN = re.compile(r"^[a-zA-Z0-9_-]+\.[a-zA-Z0-9_-]+$")


# TiDB Connector
class TiDBConnector:
    def __init__(
        self,
        database_url: Optional[str] = None,
        *,
        host: Optional[str] = None,
        port: Optional[int] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        database: Optional[str] = None,
    ):
        self.tidb_client = TiDBClient.connect(
            database_url=database_url,
            host=host,
            port=port,
            username=username,
            password=password,
            database=database,
        )
        if database_url:
            uri = MySQLDsn(database_url)
            self.host = uri.host
            self.port = uri.port
            self.username = uri.username
            self.password = uri.password
            self.database = uri.path.lstrip("/")
        else:
            self.host = host
            self.port = port
            self.username = username
            self.password = password
            self.database = database
        self.is_tidb_serverless = TIDB_SERVERLESS_HOST_PATTERN.match(self.host)

    def show_databases(self) -> list[dict]:
        return self.tidb_client.query("SHOW DATABASES").to_list()

    def switch_database(
        self,
        db_name: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        self.tidb_client = TiDBClient.connect(
            host=self.host,
            port=self.port,
            username=username or self.username,
            password=password or self.password,
            database=db_name or self.database,
        )

    def show_tables(self) -> list[str]:
        return self.tidb_client.table_names()

    def query(self, sql_stmt: str) -> list[dict]:
        return self.tidb_client.query(sql_stmt).to_list()

    def execute(self, sql_stmts: str | list[str]) -> list[dict]:
        results = []
        with self.tidb_client.session():
            if isinstance(sql_stmts, str):
                result = self.tidb_client.execute(sql_stmts)
                results.append(result.model_dump())
            elif isinstance(sql_stmts, list):
                for stmt in sql_stmts:
                    result = self.tidb_client.execute(stmt)
                    results.append(result.model_dump())
        return results

    def is_tidb_serverless(self) -> bool:
        return TIDB_SERVERLESS_HOST_PATTERN.match(self.host)

    def current_username(self) -> str:
        return self.tidb_client.query("SELECT CURRENT_USER()").scalar()

    def current_username_prefix(self) -> str:
        current_username = self.current_username()
        if TIDB_SERVERLESS_USERNAME_PATTERN.match(current_username):
            return current_username.split(".")[0]
        else:
            return current_username

    def create_user(self, username: str, password: str) -> str:
        if self.is_tidb_serverless:
            if not TIDB_SERVERLESS_USERNAME_PATTERN.match(username):
                username = f"{self.current_username_prefix()}.{username}"

        self.tidb_client.execute(
            "CREATE USER :username IDENTIFIED BY :password",
            {
                "username": username,
                "password": password,
            },
        )
        return username

    def remove_user(self, username: str) -> None:
        # Auto append the username prefix for tidb serverless.
        if self.is_tidb_serverless:
            if not TIDB_SERVERLESS_USERNAME_PATTERN.match(username):
                username = f"{self.current_username_prefix()}.{username}"

        self.tidb_client.execute(
            "DROP USER :username",
            {
                "username": username,
            },
            raise_error=True,
        )

    def disconnect(self) -> None:
        self.tidb_client.disconnect()


# App Context
@dataclass
class AppContext:
    tidb: TiDBConnector


@asynccontextmanager
async def app_lifespan(app: FastMCP) -> AsyncIterator[AppContext]:
    tidb = None
    try:
        log.info("Starting TiDB Connector...")
        tidb = TiDBConnector(
            database_url=os.getenv("TIDB_DATABASE_URL", None),
            host=os.getenv("TIDB_HOST", "127.0.0.1"),
            port=int(os.getenv("TIDB_PORT", "4000")),
            username=os.getenv("TIDB_USERNAME", "root"),
            password=os.getenv("TIDB_PASSWORD", ""),
            database=os.getenv("TIDB_DATABASE", "test"),
        )
        log.info(f"Connected to TiDB: {tidb.host}:{tidb.port}/{tidb.database}")
        yield AppContext(tidb=tidb)
    except Exception as e:
        log.error(f"Failed to connect to TiDB: {e}")
        raise e
    finally:
        if tidb:
            tidb.disconnect()


# MCP Server
mcp = FastMCP(
    "tidb",
    instructions="""You are a tidb database expert, you can help me query, create, and execute sql 
statements on the tidb database.

Notice:
- use TiDB instead of MySQL syntax for sql statements
- use `db_query("SHOW DATABASES()")` to get the current database.
- use switch_database tool only if there's explicit instruction, you can reference different databases 
via the `<db_name>.<table_name>` syntax.
- TiDB using VECTOR to store the vector data

    ```sql
    # Create a table with a vector column, HNSW index supports both `VEC_COSINE_DISTANCE` and `VEC_L2_DISTANCE` functions
    CREATE TABLE documents (
        id INT PRIMARY KEY,
        embedding VECTOR(3),
        VECTOR INDEX idx_embedding ((VEC_COSINE_DISTANCE(embedding)))
    );

    # Insert a vector into the table
    INSERT INTO documents (id, embedding) VALUES (1, '[1,2,3]');

    # Search for similar vectors using cosine similarity:

    SELECT id, document, 1 - VEC_COSINE_DISTANCE(embedding, '[1,2,3]') AS similarity
    FROM documents
    ORDER BY similarity DESC
    LIMIT 3;
    ```
    
    """,
    lifespan=app_lifespan,
)


# Tools


@mcp.tool(description="Show all databases in the tidb cluster")
def show_databases(ctx: Context) -> list[dict]:
    tidb = ctx.request_context.lifespan_context.tidb
    try:
        return tidb.show_databases()
    except Exception as e:
        log.error(f"Failed to show databases: {e}")
        raise e


@mcp.tool(
    description="""
    Switch to a specific database.
    
    Note:
    - The user has already specified the database in the configuration, so you don't need to switch 
    database before you execute the sql statements.
    """
)
def switch_database(
    ctx: Context,
    db_name: str,
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> None:
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        tidb.switch_database(db_name, username, password)
    except Exception as e:
        log.error(f"Failed to switch database to {db_name}: {e}")
        raise e


@mcp.tool(description="Show all tables in the database")
def show_tables(ctx: Context) -> list[str]:
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        return tidb.show_tables()
    except Exception as e:
        log.error(f"Failed to show tables for database {tidb.database}: {e}")
        raise e


@mcp.tool(
    description="""
    Query data from TiDB database via SQL, best practices:
    - using LIMIT for SELECT statements to avoid too many rows returned
    - using db_query to execute SELECT / SHOW / DESCRIBE / EXPLAIN ... read-only statements
    """
)
def db_query(ctx: Context, sql_stmt: str) -> list[dict]:
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        return tidb.query(sql_stmt)
    except Exception as e:
        log.error(f"Failed to execute query sql: {sql_stmt}, error: {e}")
        raise e


@mcp.tool(
    description="""
    Execute operations on TiDB database via SQL, best practices:
    - sql_stmts can be a sql statement string or a array of sql statement strings
    - using db_execute to execute INSERT / UPDATE / DELETE / CREATE / DROP ... statements
    - the sql statements will be executed in the same transaction
    """
)
def db_execute(ctx: Context, sql_stmts) -> list[dict]:
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        results = tidb.execute(sql_stmts)
        return results
    except Exception as e:
        log.error(f"Failed to execute operation sqls: sqls: {sql_stmts}, error: {e}")
        raise e


@mcp.tool(
    description="""
    Create a new database user, will return the username with prefix
    """
)
def db_create_user(ctx: Context, username: str, password: str) -> str:
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        fullname = tidb.create_user(username, password)
        return f"success, username: {fullname}"
    except Exception as e:
        log.error(f"Failed to create database user {username}: {e}")
        raise e


@mcp.tool(
    description="""
    Remove a database user in TiDB cluster.
    """
)
def db_remove_user(ctx: Context, username: str):
    tidb: TiDBConnector = ctx.request_context.lifespan_context.tidb
    try:
        tidb.remove_user(username)
        return f"success, deleted user with username {username}"
    except Exception as e:
        log.error(f"Failed to remove database user {username}: {e}")
        raise e



from pytidb.rerankers.base import BaseReranker, RerankResult
from typing import List, Optional


class LiteLLMReranker(BaseReranker):
    def __init__(
        self,
        model_name: str,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        timeout: Optional[int] = None,
    ):
        self.model_name = model_name
        self.api_key = api_key
        self.api_base = api_base
        self.timeout = timeout

    def rerank(
        self, query: str, documents: List[str], top_n: Optional[int] = None
    ) -> List[RerankResult]:
        try:
            from litellm import rerank
        except Exception:
            raise ImportError(
                "Failed to import litellm, please install it with `pip install pytidb[models]` or `pip install litellm`"
            )

        res = rerank(
            model=self.model_name,
            query=query,
            documents=documents,
            top_n=top_n,
            api_key=self.api_key,
            api_base=self.api_base,
            timeout=self.timeout,
        )

        return [
            RerankResult(index=item["index"], relevance_score=item["relevance_score"])
            for item in res.results
        ]



from typing import (
    Literal,
    Optional,
    List,
    Any,
    Dict,
    TypeVar,
    Type,
    Union,
    TYPE_CHECKING,
)

from sqlalchemy import Engine, delete, update
from sqlalchemy.orm import Session, DeclarativeMeta
from sqlmodel.main import SQLModelMetaclass
from tidb_vector.sqlalchemy import VectorAdaptor
from typing_extensions import Generic

from pytidb.base import Base
from pytidb.schema import (
    QueryBundle,
    VectorDataType,
    TableModel,
    DistanceMetric,
    ColumnInfo,
)
from pytidb.search import SearchType, SearchQuery
from pytidb.utils import (
    build_filter_clauses,
    check_text_column,
    check_vector_column,
    filter_text_columns,
    filter_vector_columns,
)

if TYPE_CHECKING:
    from pytidb import TiDBClient


T = TypeVar("T", bound=TableModel)


class Table(Generic[T]):
    def __init__(
        self,
        *,
        client: "TiDBClient",
        schema: Optional[Type[T]] = None,
        vector_column: Optional[str] = None,
        text_column: Optional[str] = None,
        distance_metric: Optional[DistanceMetric] = DistanceMetric.COSINE,
        checkfirst: bool = True,
    ):
        self._client = client
        self._db_engine = client.db_engine
        self._identifier_preparer = self._db_engine.dialect.identifier_preparer

        # Init table model.
        if type(schema) is SQLModelMetaclass:
            self._table_model = schema
        elif type(schema) is DeclarativeMeta:
            self._table_model = schema
        else:
            raise TypeError(f"Invalid schema type: {type(schema)}")
        self._columns = self._table_model.__table__.columns

        # Field for auto embedding.
        self._vector_field_configs = {}
        if hasattr(schema, "__pydantic_fields__"):
            for name, field in schema.__pydantic_fields__.items():
                # FIXME: using field custom attributes instead of it.
                if "embed_fn" in field._attributes_set:
                    embed_fn = field._attributes_set["embed_fn"]
                    source_field_name = field._attributes_set["source_field"]
                    self._vector_field_configs[name] = {
                        "embed_fn": embed_fn,
                        "vector_field": field,
                        "source_field_name": source_field_name,
                    }

        # Create table.
        self._sa_table = self._table_model.__table__
        Base.metadata.create_all(
            self._db_engine, tables=[self._sa_table], checkfirst=checkfirst
        )

        # Find vector and text columns.
        self._vector_columns = filter_vector_columns(self._columns)
        self._text_columns = filter_text_columns(self._columns)

        # Create vector index automatically.
        vector_adaptor = VectorAdaptor(self._db_engine)
        for col in self._vector_columns:
            if vector_adaptor.has_vector_index(col):
                continue
            vector_adaptor.create_vector_index(col, distance_metric)

        # Determine default vector column for vector search.
        if vector_column is not None:
            self._vector_column = check_vector_column(self._columns, vector_column)
        else:
            if len(self._vector_columns) == 1:
                self._vector_column = self._vector_columns[0]
            else:
                self._vector_column = None

        # Determine default text column for fulltext search.
        if text_column is not None:
            self._text_column = check_text_column(self._columns, text_column)
        else:
            if len(self._text_columns) == 1:
                self._text_column = self._text_columns[0]
            else:
                self._text_column = None

    @property
    def table_model(self) -> T:
        return self._table_model

    @property
    def table_name(self) -> str:
        return self._table_model.__tablename__

    @property
    def client(self) -> "TiDBClient":
        return self._client

    @property
    def db_engine(self) -> Engine:
        return self._db_engine

    @property
    def vector_column(self):
        return self._vector_column

    @property
    def vector_columns(self):
        return self._vector_columns

    @property
    def text_column(self):
        return self._text_column

    @property
    def text_columns(self):
        return self._text_columns

    @property
    def vector_field_configs(self):
        return self._vector_field_configs

    def get(self, id: Any) -> T:
        with self._client.session() as db_session:
            return db_session.get(self._table_model, id)

    def insert(self, data: T) -> T:
        # Auto embedding.
        for field_name, config in self._vector_field_configs.items():
            if getattr(data, field_name) is not None:
                # Vector embeddings is provided.
                continue

            if not hasattr(data, config["source_field_name"]):
                continue

            embedding_source = getattr(data, config["source_field_name"])
            vector_embedding = config["embed_fn"].get_source_embedding(embedding_source)
            setattr(data, field_name, vector_embedding)

        with self._client.session() as db_session:
            db_session.add(data)
            db_session.flush()
            db_session.refresh(data)
            return data

    def bulk_insert(self, data: List[T]) -> List[T]:
        # Auto embedding.
        for field_name, config in self._vector_field_configs.items():
            items_need_embedding = []
            sources_to_embedding = []

            # Skip if no embedding function is provided.
            if "embed_fn" not in config or config["embed_fn"] is None:
                continue

            for item in data:
                # Skip if vector embeddings is provided.
                if getattr(item, field_name) is not None:
                    continue

                # Skip if no source field is provided.
                if not hasattr(item, config["source_field_name"]):
                    continue

                items_need_embedding.append(item)
                embedding_source = getattr(item, config["source_field_name"])
                sources_to_embedding.append(embedding_source)

            # Batch embedding.
            vector_embeddings = config["embed_fn"].get_source_embeddings(
                sources_to_embedding
            )
            for item, embedding in zip(items_need_embedding, vector_embeddings):
                setattr(item, field_name, embedding)

        with self._client.session() as db_session:
            db_session.add_all(data)
            db_session.flush()
            for item in data:
                db_session.refresh(item)
            return data

    def update(self, values: dict, filters: Optional[Dict[str, Any]] = None) -> object:
        # Auto embedding.
        for field_name, config in self._vector_field_configs.items():
            if field_name in values:
                # Vector embeddings is provided.
                continue

            if config["source_field_name"] not in values:
                continue

            embedding_source = values[config["source_field_name"]]
            vector_embedding = config["embed_fn"].get_source_embedding(embedding_source)
            values[field_name] = vector_embedding

        with self._client.session() as db_session:
            filter_clauses = build_filter_clauses(
                filters, self._columns, self._table_model
            )
            stmt = update(self._table_model).filter(*filter_clauses).values(values)
            db_session.execute(stmt)

    def delete(self, filters: Optional[Dict[str, Any]] = None):
        """
        Delete data from the TiDB table.

        params:
            filters: (Optional[Dict[str, Any]]): The filters to apply to the delete operation.
        """
        with self._client.session() as db_session:
            filter_clauses = build_filter_clauses(
                filters, self._columns, self._table_model
            )
            stmt = delete(self._table_model).filter(*filter_clauses)
            db_session.execute(stmt)

    def truncate(self):
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = f"TRUNCATE TABLE {table_name};"
            self._client.execute(stmt)

    def columns(self) -> List[ColumnInfo]:
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = """
                SELECT
                    COLUMN_NAME as column_name,
                    COLUMN_TYPE as column_type
                FROM INFORMATION_SCHEMA.COLUMNS
                WHERE
                    TABLE_SCHEMA = DATABASE()
                    AND TABLE_NAME = :table_name;
            """
            res = self._client.query(stmt, {"table_name": table_name})
            return res.to_pydantic(ColumnInfo)

    def rows(self):
        with self._client.session():
            table_name = self._identifier_preparer.quote(self.table_name)
            stmt = f"SELECT COUNT(*) FROM {table_name};"
            return self._client.query(stmt).scalar()

    def query(self, filters: Optional[Dict[str, Any]] = None) -> List[T]:
        with Session(self._db_engine) as db_session:
            query = db_session.query(self._table_model)
            if filters:
                filter_clauses = build_filter_clauses(
                    filters, self._columns, self._table_model
                )
                query = query.filter(*filter_clauses)
            return query.all()

    def search(
        self,
        query: Optional[Union[VectorDataType, str, QueryBundle]] = None,
        search_type: SearchType = "vector",
    ) -> SearchQuery:
        return SearchQuery(
            table=self,
            query=query,
            search_type=search_type,
        )

    def _has_tiflash_index(
        self,
        column_name: str,
        index_kind: Optional[Literal["FullText", "Vector"]] = None,
    ) -> bool:
        stmt = """SELECT EXISTS(
            SELECT 1
            FROM INFORMATION_SCHEMA.TIFLASH_INDEXES
            WHERE
                TIDB_DATABASE = DATABASE()
                AND TIDB_TABLE = :table_name
                AND COLUMN_NAME = :column_name
                AND INDEX_KIND = :index_kind
        )
        """
        with self._client.session():
            res = self._client.query(
                stmt,
                {
                    "table_name": self.table_name,
                    "column_name": column_name,
                    "index_kind": index_kind,
                },
            )
            return res.scalar()

    def has_fts_index(self, column_name: str) -> bool:
        return self._has_tiflash_index(column_name, "FullText")

    def create_fts_index(self, column_name: str, name: Optional[str] = None):
        _name = self._identifier_preparer.quote(name or f"fts_idx_{column_name}")
        _column_name = self._identifier_preparer.format_column(
            self._columns[column_name]
        )

        add_tiflash_replica_stmt = (
            f"ALTER TABLE {self.table_name} SET TIFLASH REPLICA 1;"
        )
        self._client.execute(add_tiflash_replica_stmt, raise_error=True)

        create_index_stmt = (
            f"CREATE FULLTEXT INDEX {_name} ON "
            f"{self.table_name} ({_column_name}) "
            f"WITH PARSER MULTILINGUAL;"
        )
        self._client.execute(create_index_stmt, raise_error=True)



# Numeric Types
from sqlalchemy.types import BIGINT as BIGINT
from sqlalchemy.types import BigInteger as BigInteger
from sqlalchemy.types import INT as INT
from sqlalchemy.types import INTEGER as INTEGER
from sqlalchemy.types import Integer as Integer
from sqlalchemy.types import SMALLINT as SMALLINT
from sqlalchemy.types import SmallInteger as SmallInteger
from sqlalchemy.types import FLOAT as FLOAT
from sqlalchemy.types import Float as Float
from sqlalchemy.types import DOUBLE as DOUBLE
from sqlalchemy.types import Double as Double
from sqlalchemy.types import DECIMAL as DECIMAL

# String Types
from sqlalchemy.types import CHAR as CHAR
from sqlalchemy.types import VARCHAR as VARCHAR
from sqlalchemy.types import TEXT as TEXT
from sqlalchemy.types import Text as Text
from sqlalchemy.types import String as String
from sqlmodel import AutoString

# Date and Time Types
from sqlalchemy.types import DATE as DATE
from sqlalchemy.types import Date as Date
from sqlalchemy.types import DATETIME as DATETIME
from sqlalchemy.types import DateTime as DateTime
from sqlalchemy.types import TIMESTAMP as TIMESTAMP

# Boolean Types
from sqlalchemy.types import BOOLEAN as BOOLEAN
from sqlalchemy.types import Boolean as Boolean

# Binary Types
from sqlalchemy.types import BINARY as BINARY
from sqlalchemy.types import VARBINARY as VARBINARY

# JSON Type
from sqlalchemy.types import JSON as JSON

# Vector Type
from tidb_vector.sqlalchemy import VectorType as Vector
from tidb_vector.sqlalchemy import VectorType as VECTOR

# Base Type for Custom Types
from sqlalchemy.types import TypeDecorator as TypeDecorator

__all__ = [
    # Numeric Types
    "BIGINT",
    "BigInteger",
    "INT",
    "INTEGER",
    "Integer",
    "SMALLINT",
    "SmallInteger",
    "FLOAT",
    "Float",
    "DOUBLE",
    "Double",
    "DECIMAL",
    # String Types
    "CHAR",
    "VARCHAR",
    "TEXT",
    "Text",
    "String",
    "AutoString",
    # Date and Time Types
    "DATE",
    "Date",
    "DATETIME",
    "DateTime",
    "TIMESTAMP",
    # Boolean Types
    "BOOLEAN",
    "Boolean",
    # Binary Types
    "BINARY",
    "VARBINARY",
    # JSON Type
    "JSON",
    # Vector Type
    "Vector",
    "VECTOR",
    # Base Type for Custom Types
    "TypeDecorator",
]



# Basic SQL operations
from sqlmodel.sql.expression import select as select
from sqlalchemy.sql import insert as insert
from sqlalchemy.sql import update as update
from sqlalchemy.sql import delete as delete

# Table and column operations
from sqlalchemy.sql import table as table
from sqlalchemy.sql import column as column
from sqlalchemy.sql import literal as literal
from sqlalchemy.sql import literal_column as literal_column

# Join operations
from sqlalchemy.sql import join as join
from sqlalchemy.sql import outerjoin as outerjoin

# Set operations
from sqlalchemy.sql import union as union
from sqlalchemy.sql import union_all as union_all
from sqlalchemy.sql import intersect as intersect
from sqlalchemy.sql import except_ as except_

# Expression operations
from sqlmodel.sql.expression import and_ as and_
from sqlmodel.sql.expression import or_ as or_
from sqlmodel.sql.expression import not_ as not_
from sqlmodel.sql.expression import between as between
from sqlmodel.sql.expression import case as case
from sqlmodel.sql.expression import cast as cast
from sqlmodel.sql.expression import col as col
from sqlmodel.sql.expression import collate as collate
from sqlmodel.sql.expression import distinct as distinct
from sqlmodel.sql.expression import extract as extract
from sqlmodel.sql.expression import funcfilter as funcfilter
from sqlmodel.sql.expression import over as over
from sqlmodel.sql.expression import tuple_ as tuple_
from sqlmodel.sql.expression import type_coerce as type_coerce
from sqlmodel.sql.expression import within_group as within_group

# Aggregate functions
from sqlmodel.sql.expression import all_ as all_
from sqlmodel.sql.expression import any_ as any_

# Ordering and null handling
from sqlmodel.sql.expression import asc as asc
from sqlmodel.sql.expression import desc as desc
from sqlmodel.sql.expression import nulls_first as nulls_first
from sqlmodel.sql.expression import nulls_last as nulls_last

# Constants and special values
from sqlalchemy.sql import null as null
from sqlalchemy.sql import true as true
from sqlalchemy.sql import false as false
from sqlalchemy.sql import exists as exists
from sqlalchemy.sql import func as func
from sqlalchemy.sql import values as values
from sqlalchemy.sql import bindparam as bindparam
from sqlalchemy.sql import alias as alias
from sqlalchemy.sql import modifier as modifier
from sqlalchemy.sql import text as text

__all__ = [
    # Basic SQL operations
    "select",
    "insert",
    "update",
    "delete",
    # Table and column operations
    "table",
    "column",
    "literal",
    "literal_column",
    # Join operations
    "join",
    "outerjoin",
    # Set operations
    "union",
    "union_all",
    "intersect",
    "except_",
    # Expression operations
    "and_",
    "or_",
    "not_",
    "between",
    "case",
    "cast",
    "col",
    "collate",
    "distinct",
    "extract",
    "funcfilter",
    "over",
    "tuple_",
    "type_coerce",
    "within_group",
    # Aggregate functions
    "all_",
    "any_",
    # Ordering and null handling
    "asc",
    "desc",
    "nulls_first",
    "nulls_last",
    # Constants and special values
    "null",
    "true",
    "false",
    "exists",
    "func",
    "values",
    "bindparam",
    "alias",
    "modifier",
    "text",
]



from abc import ABC, abstractmethod
from typing import List, Optional

from pydantic import BaseModel


class RerankResult(BaseModel):
    index: int
    relevance_score: float


class BaseReranker(ABC):
    @abstractmethod
    def rerank(
        self,
        query: str,
        documents: List[str],
        top_n: Optional[int] = None,
    ) -> List[RerankResult]:
        raise NotImplementedError()



import os

from .client import TiDBClient
from .table import Table
from sqlmodel import Session
from sqlalchemy import create_engine

if "LITELLM_LOCAL_MODEL_COST_MAP" not in os.environ:
    os.environ["LITELLM_LOCAL_MODEL_COST_MAP"] = "True"

if "LITELLM_LOG" not in os.environ:
    os.environ["LITELLM_LOG"] = "WARNING"

__all__ = ["TiDBClient", "Table", "Session", "create_engine"]



import re

from urllib.parse import quote
from typing import Dict, Optional, Any, List, Callable, TypeVar, Tuple

import sqlalchemy
from pydantic import AnyUrl, UrlConstraints
from sqlalchemy import BinaryExpression, Column, String
from sqlmodel import AutoString
from tidb_vector.sqlalchemy import VectorType
from sqlalchemy.engine import Row
from sqlalchemy.engine.result import result_tuple
from sqlalchemy import Table
from typing import Union

from pytidb.schema import TableModel
from pytidb.constants import (
    AND,
    OR,
    EQ,
    IN,
    NIN,
    GT,
    GTE,
    LT,
    LTE,
    NE,
)

JSON_FIELD_PATTERN = re.compile(
    r"^(?P<column>[a-zA-Z_][a-zA-Z0-9_]*)\.(?P<json_field>[a-zA-Z_][a-zA-Z0-9_]*)$"
)


TIDB_SERVERLESS_HOST_PATTERN = re.compile(
    r"gateway\d{2}\.(.+)\.(prod|dev|staging)\.(shared\.)?(aws|alicloud)\.tidbcloud\.com"
)


class TiDBDsn(AnyUrl):
    """A type that will accept any TiDB DSN.

    * User info required
    * TLD not required
    * Host not required
    """

    _constraints = UrlConstraints(
        allowed_schemes=[
            "mysql",
            "mysql+mysqlconnector",
            "mysql+aiomysql",
            "mysql+asyncmy",
            "mysql+mysqldb",
            "mysql+pymysql",
            "mysql+cymysql",
            "mysql+pyodbc",
        ],
        default_port=4000,
        host_required=True,
    )


def build_tidb_dsn(
    schema: str = "mysql+pymysql",
    host: str = "localhost",
    port: int = 4000,
    username: str = "root",
    password: str = "",
    database: str = "test",
    enable_ssl: Optional[bool] = None,
) -> TiDBDsn:
    if enable_ssl is None:
        if host and TIDB_SERVERLESS_HOST_PATTERN.match(host):
            enable_ssl = True
        else:
            enable_ssl = None

    return TiDBDsn.build(
        scheme=schema,
        host=host,
        port=port,
        username=username,
        # TODO: remove quote after following issue is fixed:
        # https://github.com/pydantic/pydantic/issues/8061
        password=quote(password) if password else None,
        path=database,
        query="ssl_verify_cert=true&ssl_verify_identity=true" if enable_ssl else None,
    )


def filter_vector_columns(columns: Dict) -> List[Column]:
    vector_columns = []
    for column in columns:
        if isinstance(column.type, VectorType):
            vector_columns.append(column)
    return vector_columns


def check_vector_column(columns: Dict, column_name: str) -> Optional[str]:
    if column_name not in columns:
        raise ValueError(f"Non-exists vector column: {column_name}")

    vector_column = columns[column_name]
    if vector_column.type != VectorType:
        raise ValueError(f"Invalid vector column: {vector_column}")

    return vector_column


def filter_text_columns(columns: Dict) -> List[Column]:
    text_columns = []
    for column in columns:
        if isinstance(column.type, AutoString) or isinstance(column.type, String):
            text_columns.append(column)
    return text_columns


def check_text_column(columns: Dict, column_name: str) -> Optional[str]:
    if column_name not in columns:
        raise ValueError(f"Non-exists text column: {column_name}")

    text_column = columns[column_name]
    if not isinstance(text_column.type, String) and not isinstance(
        text_column.type, AutoString
    ):
        raise ValueError(f"Invalid text column: {text_column}")

    return text_column


def build_filter_clauses(
    filters: Dict[str, Any], columns: Dict, table_model: TableModel
) -> List[BinaryExpression]:
    if filters is None:
        return []

    filter_clauses = []
    for key, value in filters.items():
        if key.lower() == AND:
            if not isinstance(value, list):
                raise TypeError(
                    f"Expect a list value for $and operator, but got {type(value)}"
                )
            and_clauses = []
            for item in value:
                and_clauses.extend(build_filter_clauses(item, columns, table_model))
            if len(and_clauses) == 0:
                continue
            filter_clauses.append(sqlalchemy.and_(*and_clauses))
        elif key.lower() == OR:
            if not isinstance(value, list):
                raise TypeError(
                    f"Expect a list value for $or operator, but got {type(value)}"
                )
            or_clauses = []
            for item in value:
                or_clauses.extend(build_filter_clauses(item, columns, table_model))
            if len(or_clauses) == 0:
                continue
            filter_clauses.append(sqlalchemy.or_(*or_clauses))
        elif key in columns:
            column = getattr(columns, key)
            if isinstance(value, dict):
                filter_clause = build_column_filter(column, value)
            else:
                # implicit $eq operator: value maybe int / float / string
                filter_clause = build_column_filter(column, {EQ: value})
            if filter_clause is not None:
                filter_clauses.append(filter_clause)
        elif "." in key:
            match = JSON_FIELD_PATTERN.match(key)
            if not match:
                raise ValueError(
                    f"Got unexpected filter key: {key}, please use valid column name instead"
                )
            column_name = match.group("column")
            json_field = match.group("json_field")
            column = sqlalchemy.func.json_extract(
                getattr(table_model, column_name), f"$.{json_field}"
            )
            if isinstance(value, dict):
                filter_clause = build_column_filter(column, value)
            else:
                # implicit $eq operator: value maybe int / float / string
                filter_clause = build_column_filter(column, {EQ: value})
            if filter_clause is not None:
                filter_clauses.append(filter_clause)
        else:
            raise ValueError(
                f"Got unexpected filter key: {key}, please use valid column name instead"
            )

    return filter_clauses


def build_column_filter(
    column: Any, conditions: Dict[str, Any]
) -> Optional[BinaryExpression]:
    column_filters = []
    for operator, val in conditions.items():
        op = operator.lower()
        if op == IN:
            column_filters.append(column.in_(val))
        elif op == NIN:
            column_filters.append(~column.in_(val))
        elif op == GT:
            column_filters.append(column > val)
        elif op == GTE:
            column_filters.append(column >= val)
        elif op == LT:
            column_filters.append(column < val)
        elif op == LTE:
            column_filters.append(column <= val)
        elif op == NE:
            column_filters.append(column != val)
        elif op == EQ:
            column_filters.append(column == val)
        else:
            raise ValueError(
                f"Unknown filter operator {operator}. Consider using "
                "one of $in, $nin, $gt, $gte, $lt, $lte, $eq, $ne."
            )
    if len(column_filters) == 0:
        return None
    elif len(column_filters) == 1:
        return column_filters[0]
    else:
        return sqlalchemy.and_(*column_filters)


RowKeyType = TypeVar("RowKeyType", bound=Union[Any, Tuple[Any, ...]])


def get_row_id_from_row(row: Row, table: Table) -> Optional[RowKeyType]:
    pk_constraint = table.primary_key
    if not pk_constraint.columns:
        # Try to get _tidb_rowid if no primary key exists
        try:
            return row._mapping["_tidb_rowid"]
        except KeyError:
            return row.__hash__()

    pk_column_names = [col.name for col in pk_constraint.columns]
    try:
        if len(pk_column_names) == 1:
            return row._mapping[pk_column_names[0]]
        return tuple(row._mapping[name] for name in pk_column_names)
    except KeyError as e:
        raise KeyError(
            f"Primary key column '{e.args[0]}' not found in Row. "
            f"Available: {list(row._mapping.keys())}"
        )


def merge_result_rows(
    rows_a: List[Row],
    rows_b: List[Row],
    get_row_key: Callable[[Row], RowKeyType],
    merge_strategies: Optional[Dict[str, Callable[[Any, Any, Row, Row], Any]]] = None,
) -> Tuple[List[str], List[Row]]:
    """Merge two lists of result rows based on row_id.

    Args:
        rows_a: First list of result rows
        rows_b: Second list of result rows
        get_row_key: Function to get the key (primary key or _tidb_rowid) from a row
        merge_strategies: Optional dictionary mapping field names to custom merge functions.
                   Each merge function takes (value_a, value_b, row_a, row_b) as arguments
                   and returns the merged value.

    Returns:
        List of merged result rows
    """
    if not rows_a and not rows_b:
        return [], []
    if not rows_a or len(rows_a) == 0:
        return list(rows_b[0]._fields), rows_b
    if not rows_b or len(rows_b) == 0:
        return list(rows_a[0]._fields), rows_a

    # Get all column names
    fields_a = list(rows_a[0]._fields)
    fields_b = list(rows_b[0]._fields)
    all_fields = list(dict.fromkeys(fields_a + fields_b).keys())

    # Build mapping from key to row data
    rows_by_key_a = {get_row_key(row): row for row in rows_a}
    rows_by_key_b = {get_row_key(row): row for row in rows_b}

    # Get all unique keys
    all_keys = set(rows_by_key_a.keys()) | set(rows_by_key_b.keys())

    # Merge results
    merged_rows = []
    for key in all_keys:
        row_data = []
        row_a = rows_by_key_a.get(key)
        row_b = rows_by_key_b.get(key)

        for field in all_fields:
            value_a = getattr(row_a, field) if row_a and field in fields_a else None
            value_b = getattr(row_b, field) if row_b and field in fields_b else None

            # Use custom merge strategy if provided
            if merge_strategies and field in merge_strategies:
                value = merge_strategies[field](value_a, value_b, row_a, row_b)
            else:
                # Default strategy: use value_a if not None, otherwise use value_b
                value = value_a if value_a is not None else value_b

            row_data.append(value)

        # Create new Row object using result_tuple
        row_factory = result_tuple(all_fields)
        merged_row = row_factory(row_data)
        merged_rows.append(merged_row)

    return all_fields, merged_rows
